@@ -1,8 +1,10 @@
 from __future__ import annotations
 
+from warnings import warn
+
 from dataclasses import dataclass, field
 from random import random
-from typing import Dict, List, Optional, Tuple, Union
+from typing import cast, Dict, List, Optional, Tuple, Union
 
 from . import efnlp_pb2 as pb
 
@@ -72,6 +74,44 @@ class Language:
     def encode(self, s: str) -> List[int]:
         raise NotImplementedError("Base Language doesn't encode (yet)")
 
+        # General: incrementally parse forward with the longest match
+        # in a prefix tree?
+        #
+        # i = 0
+        # while i is not None:
+        #   t, i = self.encoder.encode(s, i) # token and next char (i + len(match) | None)
+        #
+        # what's an "encoder"?
+        #
+        # class Encoder:
+        #   depth: int # max(prefixes[*].depth) + 1
+        #   token: int # token to assign "here"
+        #   prefixes: Dict[str, Encoder]
+        #
+        #   def encode(self, s: str, i: int) -> Optional[Tuple[int, int]]:
+        #       if (i >= len(s)) or (s[i] not in self.prefixes): # second cond access ok
+        #           return None
+        #       # recursive encode; s[i] _is_ in prefixes
+        #       result = self.prefixes[s[i]].encode(s, i+1)
+        #       if result is None:
+        #           return (self.token, i+1) # leave it to caller to assess next step
+        #       # match found within prefixes
+        #       return result
+        #
+        # class EncoderSet:
+        #   depth: int # max(prefixes[*].depth) + 1
+        #   prefixes: Dict[str, Encoder]
+        #
+        #   def encode(self, s: str, i: int) -> Optional[Tuple[int, int]]:
+        #       if (i >= len(s)) or (s[i] not in self.prefixes): # second cond access ok
+        #           return None
+        #       return self.prefixes[s[i]].encode(s, i+1)
+        #
+        # What's a character encoder? Something like
+        #
+        #   EncoderSet(1, {c: Encoder(0, i, {}) for i, c in enumerate(ascii_letters)})
+        #
+
     def decode(self, tokens: Union[int, List[int]]) -> str:
         if isinstance(tokens, list):
             return "".join(self.ttos[t] for t in tokens)
@@ -128,12 +168,14 @@ class Sampler:
 
     @staticmethod
     def from_proto_bytes(data: bytes) -> Sampler:
+        assert isinstance(data, bytes)
         P = pb.Sampler()
         P.ParseFromString(data)
         return Sampler.from_proto(P)
 
     @staticmethod
     def from_proto(S: pb.Sampler) -> Sampler:
+        assert isinstance(S, pb.Sampler), (type(S), S)
         s = Sampler(total=S.total)
         s.counts = {stc.token: stc.count for stc in S.counts}
         return s
@@ -307,7 +349,7 @@ class SuffixTree:
 
     @staticmethod
     def merge_proto_bytes(a: bytes, b: bytes) -> SuffixTree:
-        return SuffixTree.merge_proto(
+        return SuffixTree.merge_trees(
             SuffixTree.from_proto_bytes(a), SuffixTree.from_proto_bytes(b)
         )
 
@@ -322,7 +364,6 @@ class SuffixTree:
     def merge(self, other: SuffixTree) -> SuffixTree:
         """Merge another SuffixTree into this one"""
         assert self.token == other.token
-        # merge overall target tokens and counts
         self.depth += max(self.depth, other.depth)
         self.sampler.merge(other.sampler)
         for t in self:
@@ -331,6 +372,7 @@ class SuffixTree:
         for t in other:
             if t not in self:
                 self.children[t] = SuffixTree.copy(other.children[t])
+                # TODO: is a copy required? why not just reference?
         return self
 
     def memory(self) -> int:
@@ -532,3 +574,119 @@ class SuffixTreeSet:
             prefix = code if len(code) < window else code[-window:]
             code.append(self.sample(prefix))
         return code
+
+
+AccmType = Tuple[int, bytes]
+
+
+class _SuffixTreeParser:  # (beam.DoFn):
+    def __init__(
+        self,
+        language: CharLanguage,  # is beam so specific about this?
+        block_size: int,
+        documents: bool = False,
+        doc_start_tok: int = 0,
+        doc_end_tok: int = 1,
+        text_encoded: bool = False,
+        text_delimiter: str = ",",
+        text_field: str = "segment",
+    ):
+        self.L = language
+        self.B = block_size
+
+        # TODO: annoying; setattr over generic kwargs? cleaner code
+        # but much more opaque interface/API here
+        self.documents = documents
+        self.doc_start_tok = doc_start_tok
+        self.doc_end_tok = doc_end_tok
+        self.text_encoded = text_encoded
+        self.text_delimiter = text_delimiter
+        self.text_field = text_field
+
+    def process(self, record):  # -> Generator[Tuple[int, bytes]]?
+
+        # size: int; but what about dynamic parsing?
+        T = SuffixTreeSet()
+
+        # encode segment in language; TODO: generalize
+        encoded: List[int] = []
+        text = cast(str, record[self.text_field])
+        if self.text_encoded:
+            encoded = [int(i) for i in text.split(self.text_delimiter)]
+        encoded = self.L.encode(text)
+
+        # TODO: doc start + encoded + doc end...
+        if self.documents:
+            encoded = [self.doc_start_tok] + encoded + [self.doc_end_tok]
+
+        # parse token string
+        T.parse_all(encoded, self.B)
+
+        # emit each tree
+        for t, tree in T.trees.items():
+            # TODO: tuple form sufficient for a "keyed" PCollection?
+            yield (t, bytes(tree))  # bytes -> protobuf encoding
+
+
+class _SuffixTreeMerge:  # (beam.CombineFn):
+    # are we merging based on proto, or pyobjects?
+
+    def create_accumulator(self) -> Optional[AccmType]:
+        return None  # SuffixTree() but for what token?
+
+    def add_input(self, accm: Optional[AccmType], inp: bytes) -> AccmType:
+        if accm is None:
+            tree = SuffixTree.from_proto_bytes(inp)
+            return (tree.token, bytes(tree))
+        tree = SuffixTree.merge_proto_bytes(accm[1], inp)
+        return (tree.token, bytes(tree))
+
+    # TODO: List/Iterable arg? or varargs?
+    def merge_accumulators(self, accumulators: List[Optional[AccmType]]) -> Optional[AccmType]:
+        tree = None
+        for accm in accumulators:
+            if accm:
+                t, b = accm  # split into token/bytes
+                if tree:
+                    other = SuffixTree.from_proto_bytes(b)
+                    tree.merge(other)  # won't merge if token mismatched
+                else:
+                    tree = SuffixTree.from_proto_bytes(b)
+        if tree:
+            return (tree.token, bytes(tree))
+        return None
+
+    def extract_output(self, accm: AccmType) -> AccmType:
+        return accm
+
+
+try:
+
+    import apache_beam as beam
+
+    # Note: MRO requires inheriting _SuffixTreeXXX first; o/w class will
+    # access beam.DoFn/CombineFn's process method first and error out
+
+    class SuffixTreeParser(_SuffixTreeParser, beam.DoFn):
+        pass
+
+    class SuffixTreeMerge(_SuffixTreeMerge, beam.CombineFn):
+        pass
+
+except ImportError as err:
+
+    warn(
+        f"Cannot import beam: SuffixTreeParser and SuffixTreeMerge will not be parallelized ({err})"
+    )
+
+    class Unimportable:
+        def __init__(self, *args, **kwargs) -> None:
+            raise NotImplementedError(
+                f"Cannot initialize {self.__class__.__name__} without `beam` modules"
+            )
+
+    class SuffixTreeParser(_SuffixTreeParser):  # type:ignore
+        pass
+
+    class SuffixTreeMerge(_SuffixTreeMerge):  # type:ignore
+        pass
